from typing import Any
from typing import Iterable
from typing import Optional

from report_writer.core.content.base import BaseContent
from report_writer.core.content.samples.getters import ContextContent


class ContentChain(BaseContent):
    """
    Контент описывающий получение данных как последовательное использование
    других объектов контента с передачей им результатов предыдущих шагов.

    .. code-block::

        content = ContentChain([
            KeyContent('foo'),
            TemplateContent('{{ bar }}'),
        ])
        content.render({
            'foo': {'bar': 1}
        })  # '1'

    :param content_iterator: Итератор объектов контента
    """
    def __init__(self, content_iterator: Iterable[BaseContent]) -> None:
        super(ContentChain, self).__init__()
        self._content_iterator: Iterable[BaseContent] = content_iterator

    def render(self, context: Any) -> Any:
        current_context = context

        for content in self._content_iterator:
            current_context = content.render(current_context)

        return current_context


class ContentGate(ContentChain):
    """
    Контент описывающий получение данных как последовательное использование
    2 других объектов контента.

    .. code-block::

        content = ContentGate(
            KeyContent('foo'),
            TemplateContent('{{ bar }}'),
        )
        content.render({
            'foo': {'bar': 1}
        })  # '1'

    :param enter_content: Контент "на входе"
    :param exit_content: Контента "на выходе"
    """

    default_base_content: BaseContent = ContextContent()

    def __init__(
            self,
            enter_content: Optional[BaseContent] = None,
            exit_content: Optional[BaseContent] = None
    ) -> None:
        if enter_content is None:
            enter_content = self.default_base_content

        if exit_content is None:
            exit_content = self.default_base_content

        super(ContentGate, self).__init__((enter_content, exit_content))


class ConditionContent(BaseContent):
    """
    Контент описывающий получение данных по условию.

    В зависимости от исполнения условия будет использован тот или иной
    объект контента.

    .. code-block::

        content = ConditionContent(
            (Content() > 3) & (Content() < 5),
            ContentGate(
                Content() >> (lambda i: {'context': i}),
                TemplateContent('3 < {{ context }} < 5')
            ),
            NoneContent()
        )

        content.render(3)  # None
        content.render(4)  # '3 < 4 < 5'
        content.render(5)  # None

    :param condition: Контент-условие
    :param true: Контент если условие истинно
    :param false: Контент если условие ложно
    """
    def __init__(
            self,
            condition: BaseContent,
            true: BaseContent,
            false: BaseContent
    ) -> None:
        self._condition: BaseContent = condition
        self._true: BaseContent = true
        self._false: BaseContent = false

    def render(self, context: Any) -> Any:
        if self._condition.render(context):
            return self._true.render(context)
        else:
            return self._false.render(context)
