from types import TracebackType
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

from report_writer.core.report_objects.exceptions import (
    FrameOutOfBoundUsageException
)
from report_writer.core.report_objects.exceptions import NotLastStateUsage


class CursorFrame:
    """
    Фрейм курсора.

    Фрейм представляет из себя прямоугольную область верхний левый угол которой
    фиксирован, а нижний правый выбран таким образом чтобы включать в себя все
    использованные клетки таблицы с момента открытия фрейма

    :param state: Начальное состояние курсора
    """
    def __init__(self, state: 'CursorState') -> None:
        self._start_row: int = state.cursor.row
        self._start_col: int = state.cursor.col
        self._state: CursorState = state
        self._cursor: Cursor = state.cursor
        self._has_points: bool = False
        self._max_row: int = self.start_row
        self._max_col: int = self.start_col

    @property
    def start_row(self) -> int:
        """
        :return: Начальная строка фрейма
        """
        return self._start_row

    @property
    def start_col(self) -> int:
        """
        :return: Начальная колонка фрейма
        """
        return self._start_col

    @property
    def state(self) -> 'CursorState':
        """
        :return: Состояние курсора которому принадлежит фрейм
        """
        return self._state

    @property
    def cursor(self) -> 'Cursor':
        """
        :return: Курсор которому принадлежит фрейм
        """
        return self._cursor

    @property
    def has_points(self) -> bool:
        """
        :return: Наличие точек курсора
        """
        return self._has_points

    @property
    def max_row(self) -> int:
        """
        :return: Максимальная использованная строка фрейма
        """
        return self._max_row

    @property
    def max_col(self) -> int:
        """
        :return: Максимальная использованная колонка фрейма
        """
        return self._max_col

    def reset_frame(self) -> None:
        """
        Сброс фрейма.
        """
        self._max_row = self.start_row
        self._max_col = self.start_col

    def use_point(self, row: int, col: int) -> None:
        """
        Пометить клетку фрема как использованную.

        :param row: Индекс строки
        :param col: Индекс колонки
        """
        if row < self.start_row or col < self.start_col:
            raise FrameOutOfBoundUsageException()
        self._has_points = True
        self._max_row = max(row, self.max_row)
        self._max_col = max(col, self.max_col)

    @property
    def width(self) -> int:
        """
        :return: Ширина фрейма
        """
        return self.max_col - self.start_col + 1

    @property
    def height(self) -> int:
        """
        :return: Высота фрейма
        """
        return self.max_row - self.start_row + 1

    def close(self, direction: Tuple[int, int]) -> None:
        """
        Закрытие фрейма.

        Возвращает курсор на точку начала фрейма и прыгает на конец фрейма
        (справа или снизу в зависимости от выбранного направления)

        :param direction: Направление смещения курсора после закрытия
        """
        frame_h, frame_w = self.height, self.width
        d_h, d_w = direction

        frame_h *= d_h
        frame_w *= d_w

        self.cursor.back(self.state)

        if self.has_points:
            self.cursor.move(Cursor.VERTICAL, frame_h)
            self.cursor.move(Cursor.HORIZONTAL, frame_w)


class CursorState:
    """
    Состояние курсора.

    Является менеджером контекста.

    :param cursor: Курсор которому принадлежит состояние
    """
    def __init__(self, cursor: 'Cursor') -> None:
        self._row: int = cursor.row
        self._col: int = cursor.col
        self._cursor: Cursor = cursor
        self._frame: CursorFrame = self._create_frame()

    @property
    def row(self) -> int:
        """
        :return: Строка состояния курсора
        """
        return self._row

    @property
    def col(self) -> int:
        """
        :return: Колонка состояния курсора
        """
        return self._col

    @property
    def cursor(self) -> 'Cursor':
        """
        :return: Курсор которому привязано состояние
        """
        return self._cursor

    @property
    def frame(self) -> 'CursorFrame':
        """
        :return: Фрейм курсора привязанный к состоянию
        """
        return self._frame

    def __enter__(self) -> 'CursorState':
        """
        При входе добавляем состояние в стек состояний курсора

        :return: Состояние курсора
        """
        self.cursor.add_state_point(self)
        return self

    def __exit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_value: Optional[BaseException],
            traceback: Optional[TracebackType]
    ) -> None:
        """
        При выходе достаём из стека состояние и проверяем что это текущее
        состояние
        """
        self.cursor.remove_state_point(self)

    def _create_frame(self) -> 'CursorFrame':
        """
        Создание фрейма для состояния.

        :return: Фрейм курсора для состояния
        """
        return CursorFrame(self)

    @property
    def point(self) -> Tuple[int, int]:
        """
        :return: Начальная точка состояния
        """
        return self.row, self.col


class Cursor:
    """
    Курсор.

    Объект курсора используется для координирования записи сложных объектов.

    Сложный объект, включающий в себя или использующий другие объекты,
    пробрасывает в них курсор, тем самым указывая в какую точку следует
    производить запись, внутренние компоненты в свою очередь сдвигают курсор
    на следующую позицию записи после окончания создания писателей.

    Работая с курсором следует использовать состояния курсора
    (:code:`CursorState`):

    .. code-block:: python

        with Cursor(0, 0).get_state() as state:
            ...  # Производим создание писателей и манипуляции с курсором
            # Закрываем фрейм что переместить курсор
            # на следующую свободную клетку
            state.frame.close(Cursor.VERTICAL)

    Если клетка используется для записи данных следует явно получать её через
    :code:`Cursor.use_point()` иначе клетка не будет учитываться как часть
    активного фрейма и может быть перезаписана следующими объектами отчёта.

    Курсор следует перемещать только вправо, вниз и на последнее сохранённое
    состояние, т.к. на таком поведении основана работа фреймов
    курсора (:code:`CursorFrame`).

    :param row: Индекс начальной строки курсора
    :param col: Индекс начальной колонки курсора
    """

    # Доступные направления движения курсора, крайне не рекомендуется
    # использовать другие
    HORIZONTAL: Tuple[int, int] = (0, 1)
    VERTICAL: Tuple[int, int] = (1, 0)

    def __init__(self, row: int, col: int) -> None:
        self._row: int = row
        self._col: int = col
        self._state_points: List['CursorState'] = []

    @property
    def row(self) -> int:
        """
        :return: Текущая строка курсора
        """
        return self._row

    @property
    def col(self) -> int:
        """
        :return: Текущая колонка курсора
        """
        return self._col

    def _create_state_point(self) -> 'CursorState':
        """
        Создание состояния курсора.

        :return: Состояние курсора
        """
        return CursorState(self)

    def get_state(self) -> CursorState:
        """
        Получение текущего состояния курсора.

        :return: Состояние курсора
        """
        state = self._create_state_point()
        return state

    def add_state_point(self, state: CursorState) -> None:
        """
        Добавление состояния к курсору

        :param state: Объект состояния курсора
        """
        if state.cursor is not self:
            raise ValueError('Состояние курсора не связанно с курсором')

        self._state_points.append(state)

    def remove_state_point(self, state: CursorState) -> None:
        """
        Удаление состояния у курсора

        :param state: Объект состояния курсора
        """
        if state.cursor is not self:
            raise ValueError('Состояние курсора не связанно с курсором')

        state_to_remove = self._state_points.pop()
        if state_to_remove is not state:
            self._state_points.append(state_to_remove)
            raise NotLastStateUsage()

    def back(self, state: Optional[CursorState] = None) -> None:
        """
        Возврат к последнему состоянию.
        """
        last_state = self._state_points[-1]
        if state is not None and last_state is not state:
            raise NotLastStateUsage()

        self._row = last_state.row
        self._col = last_state.col

    @property
    def point(self) -> Tuple[int, int]:
        """
        :return: Текущая точка курсора
        """
        return self.row, self.col

    def use_point(self) -> Tuple[int, int]:
        """
        Использование точки курсора.

        :return: Текущая точка курсора
        """
        for state in self._state_points:
            state.frame.use_point(*self.point)

        return self.point

    def move(self, direction: Tuple[int, int], n: int) -> None:
        """
        Передвинуть курсора в указанном направлении на n клеток.

        :param direction: Направление движения
        :param n: Расстояние
        """
        row, col = direction

        row = n * row
        col = n * col

        self._row += row
        self._col += col

    @staticmethod
    def invert_direction(direction: Tuple[int, int]) -> Tuple[int, int]:
        """
        :param direction: Направление
        :return: Противоположное направление курсора
                 (Вертикальное <-> Горизонтальное)
        """
        return direction[1], direction[0]
