from typing import Any
from typing import Tuple

from report_writer.core.report_objects.base import BaseReportObject
from report_writer.core.report_objects.cursor import Cursor
from report_writer.core.writers.base import BaseWriter


class Simple(BaseReportObject):
    """
    Элементарный объект отчёта.

    Представляет собой единичную операцию записи, использующую область
    переданной высоты и ширины.

    Может быть использовано для клетки отчёта (тогда writer - писатель
    клетки отчёта).

    :param writer: Используемый писатель
    :param height: Высота области записи
    :param width: Ширина области записи
    """
    def __init__(
            self,
            writer: BaseWriter,
            height: int = 1,
            width: int = 1,
    ) -> None:
        self._writer: BaseWriter = writer
        self._height: int = height
        self._width: int = width
        # Если ширина и высота рывны 0, то операция не использует клетки отчёта
        self._use: bool = (height, width) != (0, 0)

    @property
    def height(self) -> int:
        """
        :return: Высота области записи писателя
        """
        return self._height

    @property
    def width(self) -> int:
        """
        :return: Ширина области записи писателя
        """
        return self._width

    def write_object(
            self,
            cursor: Cursor,
            data: Any,
            direction: Tuple[int, int] = Cursor.VERTICAL,
    ) -> None:
        with cursor.get_state() as state:
            if self._use:
                row, col = cursor.use_point()
                cursor.move(Cursor.VERTICAL, self.height - 1)
                cursor.move(Cursor.HORIZONTAL, self.width - 1)
                cursor.use_point()
            else:
                row, col = cursor.point

            self._writer.write(row, col, data)
            state.frame.close(direction)


# Псевдоним для более наглядного описания клеток отчёта
# (которые так же являются частным случаем элментарного объекта отчёта)
Cell = Simple
