import abc
from typing import Any
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple

from report_writer.core.writers.base import BaseWriter


class ReportFileMock:
    """
    "Файл" отчёта.

    Вместо .xlsx файла производит запись в двумерный массив
    """
    def __init__(self, h: int, w: int) -> None:
        self.w: int = w
        self.h: int = h

        self.data: List[List[Any]] = [
            [
                None
                for _ in range(w)
            ]
            for _ in range(h)
        ]

        self.merged: List[Set[Tuple[int, int]]] = []
        self.formats: Dict[str, Any] = {}

        self._rows: Dict[int, float] = {}
        self._cols: Dict[int, float] = {}

        self.saved: bool = False

    def add_format(self, name: str, format_object: Any) -> None:
        """Добавление именованного формата."""
        self.formats[name] = format_object

    def get_format(self, name: str) -> Any:
        """Получение именованного формата."""
        return self.formats.get(name)

    def write(self, row: int, col: int, data: Any) -> None:
        """Запись в клетку."""
        self.data[row][col] = data

    def merge(
            self,
            from_row: int,
            from_col: int,
            to_row: int,
            to_col: int
    ) -> None:
        """Добавление объединения клеток."""
        r = set()
        for row in range(from_row, to_row + 1):
            for col in range(from_col, to_col + 1):
                r.add((row, col))

        self.merged.append(r)

    def set_format(self, row: int, col: int, format_: Any) -> None:
        """Смена формата клетки."""
        v = self.data[row][col]
        self.data[row][col] = (format_, v)


class ReportWriter(BaseWriter, metaclass=abc.ABCMeta):
    """
    Писатель отчёта.
    """
    def __init__(self, report: ReportFileMock) -> None:
        self.report = report


class CellWriter(ReportWriter):
    """
    Писатель клетки отчёта.
    """
    def write(self, row: int, col: int, data: Any) -> None:
        self.report.write(row, col, data)


class MergeWriter(ReportWriter):
    """
    Писатель объединения.
    """
    def __init__(
            self,
            report: ReportFileMock,
            height: int,
            width: int
    ) -> None:
        super(MergeWriter, self).__init__(report)
        self.height = height
        self.width = width

    def write(self, row: int, col: int, data: Any) -> None:
        self.report.merge(
            row, col, row + self.height - 1, col + self.width - 1)


class MergedCellWriter(CellWriter):
    """
    Писатель объединённой клетки отчёта.
    """
    def __init__(
            self,
            report: ReportFileMock,
            height: int,
            width: int
    ) -> None:
        super(MergedCellWriter, self).__init__(report)
        self.merge_writer = MergeWriter(self.report, height, width)

    def write(self, row: int, col: int, data: Any) -> None:
        self.merge_writer.write(row, col, None)
        super(MergedCellWriter, self).write(row, col, data)


class NamedFormatWriter(BaseWriter):
    """
    Писатель именованного формата.
    """
    def __init__(
            self,
            report: ReportFileMock,
            format_name: str,
            format_object: Any
    ) -> None:
        self.report = report
        self.format_name = format_name
        self.format_object = format_object

    def write(self, row: int = 0, col: int = 0, data: Any = None) -> None:
        self.report.add_format(self.format_name, self.format_object)


class FormattedCellWriter(CellWriter):
    """
    Писатель отформатированной клетки отчёта.
    """
    def __init__(self, report: ReportFileMock, format_name: str) -> None:
        super(FormattedCellWriter, self).__init__(report)
        self.format_name = format_name

    def write(self, row: int, col: int, data: Any) -> None:
        super(FormattedCellWriter, self).write(row, col, data)
        self.report.set_format(row, col,
                               self.report.get_format(self.format_name))


class FormattedMergedCellWriter(MergedCellWriter):
    """
    Писатель отформатированная объединённой клетки отчёта.
    """
    def __init__(
            self,
            report: ReportFileMock,
            height: int,
            width: int,
            format_name: str
    ) -> None:
        super(FormattedMergedCellWriter, self).__init__(report, height, width)
        self.format_name = format_name

    def write(self, row: int, col: int, data: Any) -> None:
        super(FormattedMergedCellWriter, self).write(row, col, data)
        self.report.set_format(row, col,
                               self.report.get_format(self.format_name))
